import json
import yaml

from collections.abc import Iterable, Mapping


def load_config(config, accept_lists=False, accept_dicts=True):
    if isinstance(config, str):
        with open(config, 'r', encoding='utf-8') as f:
            if config.endswith('.json'):
                return json.load(f)
            else:
                return yaml.safe_load(f)
    elif isinstance(config, Iterable) and accept_lists:
        return config
    elif isinstance(config, Mapping) and accept_dicts:
        return config
    else:
        text = 'Config should be a json/yaml filename'
        if accept_lists:
            text = text + ' or a list'
        if accept_dicts:
            text = text + ' or a dict'
        text = text + ', got "{}" instead.'.format(type(config))
        raise ValueError(text)



from datetime import datetime


class SerializedMessage:
    # a base class for e.g. logging messages
    def __init__(self, text, user_id, from_user, timestamp=None, session_id=None, **kwargs):
        self.text = text
        self.user_id = user_id
        self.from_user = from_user
        self.timestamp = timestamp or str(datetime.utcnow())
        self.session_id = session_id
        self.kwargs = kwargs
        """
        Expected kwargs:
            text
            user_id
            message_id
            from_user
            username
            reply_to_id
            source
            data        (original message in Alice)
            label       (something like intent)
            request_id  (this id the same for request and response, useful for joining logs)
        """

    def to_dict(self):
        result = {
            'text': self.text,
            'user_id': self.user_id,
            'from_user': self.from_user,
            'timestamp': self.timestamp,
            'session_id': self.session_id,
        }
        for k, v in self.kwargs.items():
            if k not in result:
                result[k] = v
        return result



from dialogic.nlu.basic_nlu import PYMORPHY


def with_number(noun, number):
    text = agree_with_number(noun=noun, number=number)
    return f'{number} {text}'


def agree_with_number(noun, number):
    last = abs(number) % 10
    tens = abs(number) % 100 // 10
    if PYMORPHY:
        parses = PYMORPHY.parse(noun)
        if parses:
            return parses[0].make_agree_with_number(abs(number)).word
    # detect conjugation based on the word ending
    if last == 1:
        return noun
    elif noun.endswith('ка'):
        if last in {2, 3, 4}:
            return noun[:-1] + 'и'
        else:
            return noun[:-1] + 'ек'
    elif noun.endswith('а'):
        if last in {2, 3, 4}:
            return noun[:-1] + 'ы'
        else:
            return noun[:-1]
    else:
        if last in {2, 3, 4}:
            return noun + 'а'
        else:
            return noun + 'ов'


def inflect_case(text, case):
    if PYMORPHY:
        res = []
        for word in text.split():
            parses = PYMORPHY.parse(word)
            word_infl = None
            if parses:
                inflected = parses[0].inflect({case})
                if inflected:
                    word_infl = inflected.word
            res.append(word_infl or word)
        return ' '.join(res)
    return text


def human_duration(hours=0, minutes=0, seconds=0):
    total = hours * 3600 + minutes * 60 + seconds
    s = total % 60
    m = (total // 60) % 60
    h = total // 3600
    parts = []
    if h:
        parts.append(with_number('час', h))
    if m:
        parts.append(with_number('минута', m))
    if s or not h and not m:
        parts.append(with_number('секунда', s))
    return ' '.join(parts)



import re

from .base import CascadableDialogManager, Context, Response
from dialogic.nlu import basic_nlu, matchers
from dialogic.utils.configuration import load_config

from typing import List, Dict, Optional


class FieldOption:
    def __init__(self, data):
        self.text = None
        self.next = None  # name of the node to jump to if this option is chosen
        if isinstance(data, str):
            self.text = data
        else:
            self.text = data['text']
            self.next = data.get('next')

    @classmethod
    def from_data(cls, data):
        if data is None:
            return None
        return cls(data)


class FieldConfig:
    def __init__(self, obj, idx=-1):
        self.validate_regexp = re.compile(obj.get('validate_regexp', '.*'))
        self.id = idx
        self.name = obj.get('name', 'field_{}'.format(idx))
        self.question = obj.get('question') or obj.get('q')
        self.repeated_question = obj.get('repeated_question')
        assert self.question, f'The question should be non-empty, but it is absent in the form {obj}!'
        self.validate_message = obj.get('validate_message')
        self.options = [FieldOption(o) for o in (obj.get('options') or [])]
        self.multivalued = obj.get('multivalued') or False
        self.exit_option = FieldOption.from_data(obj.get('exit_option'))
        self.suggests = obj.get('suggests')
        self.next = obj.get('next')
        self.validate_raw = obj.get('validate_raw', False)

        if self.options is not None:
            self.matcher = matchers.make_matcher(**obj.get('matching', {'key': 'levenshtein', 'threshold': 0.8}))
            self.matcher.fit(self.all_options_texts, self.all_options_texts)
        else:
            self.matcher = None

    @property
    def all_options(self) -> List[FieldOption]:
        if not self.options:
            return []
        result = self.options
        if self.exit_option:
            return result + [self.exit_option]
        return result

    @property
    def all_options_texts(self) -> List[str]:
        return [s.text for s in self.all_options]

    def update_value(self, text, old_value=None):
        if self.multivalued:
            if old_value is None:
                old_value = []
            return old_value[:] + [text]
        return text

    def get_next_name(self, response_text=None, matched_id=None) -> Optional[str]:
        if matched_id is None and response_text in self.all_options_texts:
            matched_id = self.all_options_texts.index(response_text)
        if matched_id is not None:
            o = self.all_options[matched_id]
            if o.next:
                return o.next
        if self.next:
            return self.next


class FormConfig:
    def __init__(self, config):
        self._cfg = load_config(config)
        # todo: validate everything
        self.form_name = self._cfg['form_name']

        self.start_regex = re.compile(self._cfg['start'].get('regexp', '.*'))
        self.start_message = self._cfg['start'].get('message')
        self.start_suggests = self._cfg['start'].get('suggests', [])

        exit_block = self._cfg.get('exit', {})
        self.exit_regexp = re.compile(exit_block['regexp']) if 'regexp' in exit_block else None
        self.exit_message = exit_block.get('message')
        self.exit_suggest = exit_block.get('suggest')

        self.finish_message = self._cfg.get('finish', {}).get('message')
        self.finish_suggests = self._cfg.get('finish', {}).get('suggests')

        self.fields: List[FieldConfig] = [FieldConfig(obj, idx=i) for i, obj in enumerate(self._cfg['fields'])]
        self.name2field: Dict[str, FieldConfig] = {c.name: c for c in self.fields}
        self.default_field = FieldConfig(self._cfg['default_field']) if 'default_field' in self._cfg else None
        self.num_fields = len(self.fields)


class FormFillingDialogManager(CascadableDialogManager):
    def __init__(self, config, *args, **kwargs):
        super(FormFillingDialogManager, self).__init__(*args, **kwargs)
        self.config = FormConfig(config)

    def try_to_respond(self, ctx: Context):
        user_object = ctx.user_object or {}
        raw = ctx.message_text
        normalized = basic_nlu.fast_normalize(raw)
        form = user_object.get('forms', {}).get(self.config.form_name, {})
        form['name'] = self.config.form_name
        if form.get('is_active'):
            if self.config.exit_regexp and re.match(self.config.exit_regexp, normalized):
                form['is_active'] = False
                return Response(text=self.config.exit_message, user_object=user_object)
            question_id = form['next_question']
            validated_answer = self.validate_answer(form, ctx.message_text)
            if validated_answer is not None:
                q = self.config.fields[question_id]
                form['fields'][q.name] = q.update_value(text=validated_answer, old_value=form['fields'].get(q.name))
                next_question_id = question_id
                if not q.exit_option or q.exit_option.text == validated_answer:
                    next_question_id += 1
                next_name = q.get_next_name(response_text=validated_answer)
                if next_name:
                    next_question_id = self.config.name2field[next_name].id
                if next_question_id >= self.config.num_fields:
                    form.pop('next_question')
                    form['is_active'] = False
                    result = self.handle_completed_form(form, user_object, ctx)
                    if result is not None:
                        return result
                    return Response(
                        text=self.config.finish_message,
                        user_object=user_object,
                        suggests=self.config.finish_suggests or []
                    )
                form['next_question'] = next_question_id
                return self.ask_question(next_question_id, user_object=user_object, reask=False, form=form)
            else:
                return self.ask_question(question_id, user_object=user_object, reask=True, form=form)
        elif re.match(self.config.start_regex, normalized):
            return self.start_dialogue(ctx)
        return None

    def start_dialogue(self, ctx: Context):
        """ Initialize the form, and say the intro or ask the first question.
        This function is taken out of try_to_respond, so that it could be triggered by some external factor.
        """
        user_object = ctx.user_object or {}
        if 'forms' not in user_object:
            user_object['forms'] = {}
        user_object['forms'][self.config.form_name] = {
            'fields': {},
            'is_active': True,
            'name': self.config.form_name,
            'next_question': -1
        }
        form = user_object['forms'][self.config.form_name]
        if self.config.start_message is not None:
            form['next_question'] = -1
            return Response(
                text=self.config.start_message, suggests=self.config.start_suggests, user_object=user_object
            )
        else:
            form['next_question'] = 0
            return self.ask_question(0, user_object, reask=False)

    def ask_question(self, next_question_id, user_object, reask=False, form=None):
        the_question = self.config.fields[next_question_id]
        form = form or {}
        fields = form.get('fields') or {}
        prev_value = fields.get(the_question.name)
        response = the_question.question
        if reask:
            if the_question.validate_message is not None:
                response = the_question.validate_message
            elif self.config.default_field and 'validate_message' in self.config.default_field:
                response = self.config.default_field['validate_message']
        options = the_question.all_options_texts
        if options:
            suggests = options[:]
            if the_question.multivalued and prev_value:
                suggests = [s for s in suggests if s not in prev_value]
                if the_question.repeated_question and not reask:
                    response = the_question.repeated_question
        elif the_question.suggests:
            suggests = the_question.suggests[:]
        else:
            suggests = []
        if self.config.exit_suggest is not None:
            suggests.append(self.config.exit_suggest)
        return Response(user_object=user_object, text=response, suggests=suggests)

    def validate_answer(self, form, message_text):
        question_id = form.get('next_question', -1)
        if question_id == -1:
            return message_text
        the_question = self.config.fields[question_id]

        if the_question.validate_raw:
            normalized_text = message_text
        else:
            normalized_text = basic_nlu.fast_normalize(message_text)

        if the_question.all_options:
            winner_label, best_score = the_question.matcher.match(message_text)
            return winner_label

        if the_question.validate_regexp is not None:
            if re.match(the_question.validate_regexp, normalized_text):
                return message_text
            else:
                return None

        return message_text

    def handle_completed_form(self, form, user_object, ctx):
        """ This method can be overwritten to do something useful.
        If it is not overwritten, then the dialog ends by the `finish_message` from config.
        If it is overwritten, then it should probably put the updated `user_object` argument into its Response.
        """
        return None



from .dialog import Context
from .dialog.names import REQUEST_TYPES


def is_morning_show_context(ctx: Context) -> bool:
    if not ctx.yandex or not ctx.yandex.request:
        return False
    r = ctx.yandex.request
    if r.type != REQUEST_TYPES.SHOW_PULL:
        return False
    return r.show_type == 'MORNING'



import random
import yaml

from collections.abc import Iterable

from ..nlu import basic_nlu
from ..nlu.matchers import make_matcher
from .base import CascadableDialogManager, Context, Response


class FAQDialogManager(CascadableDialogManager):
    """ This dialog manager tries to match the input message with one of the questions from its config,
    and if successful, gives the corresponding answer. """
    def __init__(self, config, matcher='tf-idf', *args, **kwargs):
        super(FAQDialogManager, self).__init__(*args, **kwargs)
        if isinstance(config, str):
            with open(config, 'r', encoding='utf-8') as f:
                self._cfg = yaml.safe_load(f)
        elif isinstance(config, Iterable):
            self._cfg = config
        else:
            raise ValueError('Config must be a filename or a list.')
        if isinstance(matcher, str):
            matcher = make_matcher(matcher)
        self.matcher = matcher
        self._q2i = {}
        self._i2a = {}
        self._i2s = {}
        question_keys = []
        question_labels = []
        for i, pair in enumerate(self._cfg):
            questions = self._extract_string_or_strings(pair, key='q')
            for q in questions:
                q2 = self._normalize(q)
                self._q2i[q2] = i
                question_keys.append(q2)
                question_labels.append(i)
            self._i2a[i] = self._extract_string_or_strings(pair, key='a')
            self._i2s[i] = self._extract_string_or_strings(pair, key='s', allow_empty=True)
        self.matcher.fit(question_keys, question_labels)

    def try_to_respond(self, ctx: Context):
        text = self._normalize(ctx.message_text)
        index, score = self.matcher.match(text)
        if index is None:
            return None
        response = random.choice(self._i2a[index])
        suggests = self._i2s.get(index, [])
        return Response(text=response, suggests=suggests, user_object=ctx.user_object).set_text(response)

    @staticmethod
    def _extract_string_or_strings(data, key, allow_empty=False):
        if key not in data:
            if allow_empty:
                return []
            raise ValueError('The question "{}" has no "{}" key.'.format(data, key))
        inputs = data[key]
        if isinstance(inputs, str):
            result = [inputs]
        elif isinstance(inputs, Iterable):
            if not all(isinstance(i, str) for i in inputs):
                raise ValueError('The list "{}" does not consist of strings.'.format(inputs))
            result = inputs
        else:
            raise ValueError('The question "{}" is not a string or list.'.format(data))
        return result

    def _normalize(self, text):
        return basic_nlu.fast_normalize(text)



from dialogic.cascade import Cascade, DialogTurn
from dialogic.dialog_manager import TurnDialogManager
from dialogic.testing.testing_utils import make_context


def test_turn_dm():
    csc = Cascade()

    @csc.add_handler(priority=0)
    def fallback(turn: DialogTurn):
        turn.response_text = 'hi'

    @csc.add_handler(priority=1, intents=['shalom'])
    def fallback(turn: DialogTurn):
        turn.response_text = 'shalom my friend'

    dm = TurnDialogManager(cascade=csc, intents_file='tests/test_managers/intents.yaml')
    ctx = make_context('shalom')
    resp = dm.respond(ctx)
    assert resp.text == 'shalom my friend'



import re

import attr
import logging
import math

from typing import Callable, List, Optional, Dict

from .turn import DialogTurn


logger = logging.getLogger(__name__)

CHECKER_TYPE = Callable[[DialogTurn], bool]
HANDLER_TYPE = Callable[[DialogTurn], bool]
POSTPROCESSOR_TYPE = Callable[[DialogTurn, Optional[Dict]], None]


class Pr:
    EPSILON = 1e-5
    CRITICAL = 100
    STAGE = 10
    STRONG_INTENT = 5  # like 'cancel' for multi-turn scenarios
    WEAK_STAGE = 3  # stages in multi-turn scenarios that can be easily cancelled
    CHECKER = 2
    INTENT_PLUS = 1.1  # local intents, that should compete with each other but not with Yandex intents
    INTENT = 1
    INTENT_MINUS = 0.9
    FAQ = 0.1
    FIND_ANYTHING = 0.05
    BEFORE_FALLBACK = 0.03
    FALLBACK = 0.01


@attr.s
class CascadeItem:
    handler: HANDLER_TYPE = attr.ib()
    priority: float = attr.ib()
    intents: Optional[List[str]] = attr.ib(factory=list)
    stages: Optional[List[str]] = attr.ib(factory=list)
    checker: Optional[CHECKER_TYPE] = attr.ib(default=None)
    regexp: Optional[str] = attr.ib(default=None)


class Cascade:
    def __init__(self):
        self.handlers: List[CascadeItem] = []
        self.postprocessors: Dict[str, POSTPROCESSOR_TYPE] = {}

    def add_handler(
            self,
            priority=None,
            intents=None,
            stages=None,
            checker=None,
            regexp=None,
    ) -> Callable[[HANDLER_TYPE], HANDLER_TYPE]:
        if priority is None:
            if stages:
                priority = Pr.STAGE
            elif intents or regexp:
                priority = Pr.INTENT
            elif checker:
                priority = Pr.CHECKER
            else:
                priority = Pr.FALLBACK

        def wrap(f: HANDLER_TYPE):
            self.handlers.append(CascadeItem(
                handler=f,
                priority=priority,
                intents=intents,
                stages=stages,
                checker=checker,
                regexp=regexp
            ))
            return f
        return wrap

    def __call__(self, turn: DialogTurn) -> Optional[str]:
        if turn.is_complete:
            return None
        candidates = []
        for item in self.handlers:
            # stages are matched strictly
            if item.stages:
                if turn.prev_stage not in item.stages:
                    continue
            # intent scores are matched strictly and then sorted
            intent_score = -math.inf
            if item.intents:
                intent_score = max(turn.intents.get(intent, -math.inf) for intent in item.intents)
            if intent_score < 1 and item.regexp and turn.text is not None:
                if re.match(item.regexp, turn.text):
                    intent_score = 1
            if (item.intents or item.regexp) and intent_score == -math.inf:
                continue
            candidates.append((item.priority, intent_score, item))
        candidates.sort(key=lambda x: x[:2], reverse=True)
        logger.debug('sorted candidates: {}'.format([c[2].handler.__name__ for c in candidates]))

        for candidate in candidates:
            item = candidate[2]
            if item.checker and not item.checker(turn):
                continue
            result = item.handler(turn)
            if turn.is_complete:
                return item.handler.__name__
        return None

    def add_postprocessor(self, name: str, function: POSTPROCESSOR_TYPE):
        self.postprocessors[name] = function

    def get_postprocessor(self, name: str) -> Optional[POSTPROCESSOR_TYPE]:
        return self.postprocessors.get(name)

    def postprocessor(self, function: POSTPROCESSOR_TYPE):
        # this will be used as a decorator
        name = function.__name__
        if name in self.postprocessors:
            logger.warning(
                'registering postprocessor "{}" for a second time'.format(name)
            )
        self.add_postprocessor(name, function)
        return function

    def postprocess(self, turn: DialogTurn):
        if turn.is_complete and turn.can_take_control and turn.agenda:
            key, form = turn.pop_agenda()
            logger.debug('agenda key is: {}'.format(key))
            if not key:
                return
            f = self.get_postprocessor(key)
            if not f:
                return
            if form:
                f(turn, form=form)
            else:
                f(turn)



from unittest.mock import patch, MagicMock

from dialogic.interfaces.vk import VKBot, VKMessage


@patch('requests.post')
def test_send_message(mock_post: MagicMock):
    mock_post.return_value.status_code = 200
    mock_post.return_value.json = lambda: {}

    bot = VKBot(token='12345', group_id=12345)
    bot.send_message(peer_id=666, text='hello', keyboard={'buttons': [[{'action': {'type': 'text', 'label': 'yay'}}]]})
    assert mock_post.called
    args, kwargs = mock_post.call_args
    assert kwargs['url'].endswith('messages.send')
    assert kwargs['data']['peer_id'] == 666
    assert kwargs['data']['message'] == 'hello'
    assert kwargs['data']['access_token'] == '12345'
    assert 'group_id' not in kwargs['data']
    assert isinstance(kwargs['data']['keyboard'], str)


def test_apply_handlers():
    bot = VKBot(token='12345', group_id=12345)
    processed1 = []
    processed2 = []
    processed3 = []
    processed4 = []

    @bot.message_handler(regexp='hello')
    def handle1(message: VKMessage):
        processed1.append(message)

    @bot.message_handler(types=['strange_type'])
    def handle2(message: VKMessage):
        processed2.append(message)

    @bot.message_handler(func=lambda x: len(x['object']['message']['text']) == 3)
    def handle3(message: VKMessage):
        processed3.append(message)

    @bot.message_handler()
    def handle4(message: VKMessage):
        processed4.append(message)

    assert len(bot.message_handlers) == 4

    bot.process_new_updates([{'type': 'message_new', 'object': {'message': {'text': 'hello bot'}}}])
    assert len(processed1) == 1
    assert len(processed2) == 0
    assert processed1[0].text == 'hello bot'

    bot.process_new_updates([{'type': 'strange_type', 'object': {'message': {'text': 'hello bot'}}}])
    assert len(processed1) == 1
    assert len(processed2) == 1

    bot.process_new_updates([{'type': 'message_new', 'object': {'message': {'text': 'wow'}}}])
    assert len(processed3) == 1

    bot.process_new_updates([{'type': 'message_new', 'object': {'message': {'text': 'fallback'}}}])
    assert len(processed4) == 1


def test_webhook_processing():
    bot = VKBot(token='12345', group_id=12345)
    bot.webhook_key = 'secret'
    assert bot.process_webhook_data({'type': 'confirmation', 'group_id': 12345}) == ('secret', 200)

    messages = []

    @bot.message_handler()
    def handle(message):
        messages.append(message)

    new_message = {'type': 'message_new', 'object': {'message': {'text': 'hello bot'}}}
    assert bot.process_webhook_data(new_message) == ('ok', 200)
    assert messages[0].text == 'hello bot'


def test_polling():
    bot = VKBot(token='12345', group_id=12345)

    # test setting polling server
    assert bot._polling_server is None
    with patch('requests.get') as mock_get:
        mock_get.return_value.status_code = 200
        mock_get.return_value.json = lambda: {'response': {'server': 'abcd', 'key': 'xyz', 'ts': 23}}
        bot.set_polling_server()
        assert mock_get.called
        assert mock_get.call_args[1]['url'].endswith('groups.getLongPollServer')
        assert bot._polling_server == 'abcd'

    # test actually polling
    with patch('requests.get') as mock_get:
        mock_get.return_value.status_code = 200
        mock_get.return_value.json = lambda: {'updates': ['i am an update'], 'ts': 38}
        assert bot.retrieve_updates() == ['i am an update']
        assert mock_get.called
        assert mock_get.call_args[1]['url'] == 'abcd'
        assert mock_get.call_args[1]['params']['key'] == 'xyz'
        assert mock_get.call_args[1]['params']['ts'] == 23
        assert bot._polling_ts == 38


def test_webhook_remove():
    bot = VKBot(token='12345', group_id=12345)
    with patch('requests.get') as mock_get:
        mock_get.return_value.status_code = 200
        mock_get.return_value.json = lambda: {'response': {'items': [{'id': 17}, {'id': 18}]}}
        bot.remove_webhook()
        assert mock_get.call_count == 3
        assert mock_get.call_args_list[0][1]['url'].endswith('groups.getCallbackServers')
        for call_args, item_id in zip(mock_get.call_args_list[1:], [17, 18]):
            assert call_args[1]['url'].endswith('groups.deleteCallbackServer')
            assert call_args[1]['params']['server_id'] == item_id


def test_webhook_set():
    bot = VKBot(token='12345', group_id=12345)
    assert bot.webhook_key is None
    with patch('requests.get') as mock_get:
        mock_get.return_value.status_code = 200
        mock_get.return_value.json = lambda: {'response': {'code': '777', 'server_id': 13}}
        bot.set_webhook(url='localhost:15777', remove_old=False)
        assert bot.webhook_key == '777'
        assert mock_get.call_count == 3
        assert mock_get.call_args_list[0][1]['url'].endswith('groups.getCallbackConfirmationCode')

        assert mock_get.call_args_list[1][1]['url'].endswith('groups.addCallbackServer')
        assert mock_get.call_args_list[1][1]['params']['secret_key'] == '777'
        assert mock_get.call_args_list[1][1]['params']['url'] == 'localhost:15777'

        assert mock_get.call_args_list[2][1]['url'].endswith('groups.setCallbackSettings')
        assert mock_get.call_args_list[2][1]['params']['server_id'] == 13



import copy
import logging

from datetime import datetime

from ..dialog.serialized_message import SerializedMessage
from dialogic.dialog import Context, Response
from dialogic.dialog.names import SOURCES, REQUEST_TYPES
from dialogic.storage.database_utils import get_mongo_or_mock, fix_bson_keys


try:
    import pymongo
except ModuleNotFoundError:
    pymongo = None


logger = logging.getLogger(__name__)


class BaseMessageLogger:
    def __init__(self, detect_pings=False, not_log_id=None, ignore_show=True):
        self.detect_pings = detect_pings
        self.not_log_id = not_log_id or set()
        self.ignore_show = ignore_show

    def log_context(self, context, **kwargs):
        raise DeprecationWarning(
            'This operation is no longer supported - please, use log_data method.'
        )

    def log_response(self, data, source, context=None, response=None, **kwargs):
        raise DeprecationWarning(
            'This operation is no longer supported - please, use log_data method.'
        )

    def log_message(self, message, source, context=None, response=None, **kwargs):
        raise DeprecationWarning(
            'This operation is no longer supported - please, use log_data method.'
        )

    def log_data(
            self,
            data: SerializedMessage,
            context: Context = None,
            response: Response = None,
            **kwargs
    ):
        if not data:
            return
        if response is not None and response.label is not None:
            data.kwargs['label'] = response.label
        if self.should_ignore_message(result=data, context=context, response=response):
            return
        self.save_a_message(data.to_dict())

    def is_like_ping(self, context=None):
        return context is not None and context.source == SOURCES.ALICE \
               and context.message_text == 'ping' and context.session_is_new()

    def should_ignore_message(
            self, result: SerializedMessage, context: Context = None, response: Response = None
    ) -> bool:
        if self.not_log_id is not None and result.user_id in self.not_log_id:
            # main reason: don't log pings from Yandex
            return True
        if self.detect_pings and self.is_like_ping(context):
            return True
        if self.ignore_show:
            if context.yandex and context.yandex.request and context.yandex.request.type == REQUEST_TYPES.SHOW_PULL:
                return True
        return False

    def save_a_message(self, message_dict):
        logger.warning('You are using a BaseMessageLogger that does not store messages. '
                       'Please extend it to save logs directly to a database.')
        logger.info(message_dict)


class MongoMessageLogger(BaseMessageLogger):
    def __init__(self, collection=None, database=None, collection_name='message_logs', write_concern=0, **kwargs):
        super(MongoMessageLogger, self).__init__(**kwargs)
        self.collection = collection
        if self.collection is None:
            if database is None:
                database = get_mongo_or_mock()
            if pymongo and not isinstance(write_concern, pymongo.write_concern.WriteConcern):
                write_concern = pymongo.write_concern.WriteConcern(w=write_concern)
            self.collection = database.get_collection(collection_name, write_concern=write_concern)

    def save_a_message(self, message_dict):
        self.collection.insert_one(fix_bson_keys(message_dict))



import re

from typing import Dict

from ..adapters.base import BaseAdapter, Context, Response
from ..dialog.names import SOURCES, COMMANDS


class SalutAdapter(BaseAdapter):
    SOURCE = SOURCES.SALUT

    def __init__(self, split_bubbles='\n\n', **kwargs):
        super(SalutAdapter, self).__init__(**kwargs)
        self.split_bubbles = split_bubbles

    def make_context(self, message: Dict, **kwargs) -> Context:
        metadata = {}

        user = message.get('uuid')
        payload = message.get('payload') or {}
        pm = payload.get('message')

        user_id = self.SOURCE + '__' + user.get('userId')

        message_text = (pm or {}).get('original_text') or ''

        metadata['new_session'] = payload.get('new_session')

        ctx = Context(
            user_object=None,
            message_text=message_text,
            metadata=metadata,
            user_id=user_id,
            session_id=message.get('sessionId'),
            source=self.SOURCE,
            raw_message=message,
        )

        # ctx.request_type = message['request'].get('type', REQUEST_TYPES.SIMPLE_UTTERANCE)
        # ctx.payload = message['request'].get('payload', {})

        # todo: add a structured Salut request
        return ctx

    def make_response(self, response: Response, original_message=None, **kwargs):
        original_message = original_message or {}
        original_payload = original_message.get('payload') or {}

        items = []

        if response.text:
            if self.split_bubbles:
                texts = response.text.split(self.split_bubbles)
            else:
                texts = [response.text]
            for t in texts:
                items.append({'bubble': {'text': t}})

        buttons = []
        if response.links:
            for s in response.links:
                buttons.append({'title': s['title'], 'action': {'deep_link': s['url'], 'type': 'deep_link'}})
        if response.suggests:
            for s in response.suggests:
                buttons.append({'title': s, 'action': {'text': s, 'type': 'text'}})

        # todo: add cards

        if response.voice:
            voice = re.sub('<.*>', ' ', response.voice).strip().replace('\n', ' ')
        else:
            voice = response.text

        payload = {
            'pronounceText': voice,
            # pronounceTextType
            # emotion
            'items': items,
            'device': original_payload.get('device'),
            'auto_listening': True,
            'finished': bool(COMMANDS.EXIT in response.commands),
            # intent
            # asr_hints
        }
        if buttons:
            payload['suggestions'] = {'buttons': buttons}

        result = {
            "messageName": "ANSWER_TO_USER",
            "sessionId": original_message.get('sessionId'),
            "messageId": original_message.get('messageId'),
            "uuid": original_message.get('uuid'),
            "payload": payload
        }

        return result



import pytest

from dialogic import COMMANDS
from dialogic.adapters import SalutAdapter
from dialogic.dialog_connector import DialogConnector, SOURCES
from dialogic.dialog_manager import BaseDialogManager, Response


class Repeater(BaseDialogManager):
    def respond(self, ctx):
        return Response(text=ctx.message_text)


@pytest.fixture
def alice_message():
    return {
        'meta': {
            'client_id': 'ru.yandex.searchplugin/7.16 (none none; android 4.4.2)',
            'interfaces': {
                'account_linking': {},
                'payments': {},
                'screen': {}
            },
            'locale': 'ru-RU',
            'timezone': 'UTC'
        },
        'request': {
            'command': 'привет',
            'nlu': {
                'entities': [],
                'tokens': ['привет']
            },
            'original_utterance': 'передай тестовому навыку привет',
            'type': 'SimpleUtterance'
        },
        'session': {
            'message_id': 0,
            'new': True,
            'session_id': 'session-1',
            'skill_id': 'skill-1',
            'user_id': 'user-id-1'
        },
        'version': '1.0'
    }


def test_alice_input(alice_message):
    connector = DialogConnector(Repeater())
    ctx = connector.make_context(message=alice_message, source=SOURCES.ALICE)
    assert ctx.user_id == 'alice__user-id-1'
    assert ctx.message_text == 'привет'
    assert ctx.metadata['new_session']


def test_alice_response(alice_message):
    connector = DialogConnector(Repeater())
    response = Response(text='привет', suggests=['тебе тоже'])
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert not output['response'].get('end_session')
    assert output['response']['text'] == 'привет'
    assert 'voice' not in output['response'] or output['response']['voice'] == 'привет'
    assert output['response']['buttons'] == [{'title': 'тебе тоже', 'hide': True}]


def test_user_objects():
    connector = DialogConnector(Repeater())
    assert connector.get_user_object('user_1') == {}
    connector.set_user_object('user_1', {'name': 'alex'})
    assert connector.get_user_object('user_2') == {}
    assert connector.get_user_object('user_1') == {'name': 'alex'}


def test_save_user_object_by_yandex(alice_message):
    connector = DialogConnector(Repeater(), alice_native_state=True)
    uo = {
        'session': {'count': 1},
        'user': {'name': 'Alex'}
    }
    response = Response(text='привет', user_object=uo)
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert output['session_state'] == {'count': 1}
    assert output['user_state_update'] == {'name': 'Alex'}


def test_save_user_object_by_yandex_to_session(alice_message):
    connector = DialogConnector(Repeater(), alice_native_state='session')
    uo = {'count': 1, 'name': 'Alex'}
    response = Response(text='привет', user_object=uo)
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert output['session_state'] == {'count': 1, 'name': 'Alex'}


def test_save_user_object_by_yandex_to_user(alice_message):
    connector = DialogConnector(Repeater(), alice_native_state='user')
    alice_message['session']['user'] = {'user_id': '123'}
    uo = {'count': 1, 'name': 'Alex'}
    response = Response(text='привет', user_object=uo)
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert output['user_state_update'] == {'count': 1, 'name': 'Alex'}
    assert output.get('application_state') is None


def test_save_user_object_by_yandex_to_user_unauthorized(alice_message):
    connector = DialogConnector(Repeater(), alice_native_state='user')
    uo = {'count': 1, 'name': 'Alex'}
    response = Response(text='привет', user_object=uo)
    output = connector.standardize_output(source=SOURCES.ALICE, original_message=alice_message, response=response)
    assert output['user_state_update'] == {'count': 1, 'name': 'Alex'}
    assert output['application_state'] == {'count': 1, 'name': 'Alex'}


@pytest.mark.parametrize('ans,uo,auth', [
    (True, {
        'session': {'value': 10},
        'user': {'value': 42},
        'application': {'value': 50}
    }, True),
    ('session', {'value': 10}, True),
    ('user', {'value': 42}, True),
    ('user', {'value': 50}, False),
    (False, {}, True),
])
def test_load_user_object_by_yandex(ans, uo, auth, alice_message):
    connector = DialogConnector(Repeater(), alice_native_state=ans)
    if auth:
        alice_message['session']['user'] = {'user_id': '123'}
    alice_message['state'] = {
        'session': {
            'value': 10
        },
        'user': {
            'value': 42,
        },
        'application': {
            'value': 50,
        }
    }
    ctx = connector.make_context(message=alice_message, source=SOURCES.ALICE)
    assert ctx.user_object == uo


@pytest.fixture
def salut_req():
    result = {
        'messageId': 1614362356647,
        'sessionId': 'a298a38a-faa8-442f-97fd-1a53555c6d26',
        'messageName': 'MESSAGE_TO_SKILL',
        'payload': {
            'applicationId': '975dd918-8833-42ff-8c4f-d97dadb3368f',
            'appversionId': 'afda44e3-72dd-4d63-9cf6-3e6371e0d775',
            'projectName': 'b64d2304-20ea-4bc5-8e82-0b194bf79be8',
            'intent': 'run_app',
            'original_intent': 'axon',
            'intent_meta': {},
            'message': {
                'original_text': 'настоящий',
                'normalized_text': 'настоящий .',
                'tokenized_elements_list': [
                    {
                        'text': 'настоящий',
                        'raw_text': 'настоящий',
                        'grammem_info': {
                            'case': 'nom',
                            'degree': 'pos',
                            'gender': 'masc',
                            'number': 'sing',
                            'raw_gram_info': 'case=nom|degree=pos|gender=masc|number=sing',
                            'part_of_speech': 'ADJ'
                        },
                        'lemma': 'настоящий',
                        'is_stop_word': False,
                        'list_of_dependents': [],
                        'dependency_type': 'root',
                        'head': 0
                    },
                    {
                        'raw_text': '.',
                        'text': '.',
                        'lemma': '.',
                        'token_type': 'SENTENCE_ENDPOINT_TOKEN',
                        'token_value': {'value': '.'},
                        'list_of_token_types_data': [
                            {
                                'token_type': 'SENTENCE_ENDPOINT_TOKEN',
                                'token_value': {
                                    'value': '.'}}]
                    }],
                'entities': {},
                'original_message_name': 'MESSAGE_FROM_USER',
                'human_normalized_text': 'настоящий',
                'asr_normalized_message': None,
                'human_normalized_text_with_anaphora': 'настоящий'},
            'device': {'platformType': 'web', 'platformVersion': '1',
                       'surface': 'SBERBOX', 'surfaceVersion': '1',
                       'deviceId': '', 'deviceManufacturer': '',
                       'deviceModel': ''},
            'app_info': {
                'projectId': 'e4f4a982-f80d-47cc-89d2-3321fe2b11fd',
                'applicationId': '975dd918-8833-42ff-8c4f-d97dadb3368f',
                'appversionId': 'afda44e3-72dd-4d63-9cf6-3e6371e0d775',
                'systemName': None,
                'frontendEndpoint': None,
                'frontendType': 'DIALOG'
            }, 'annotations': {
                'censor_data': {'classes': ['politicians', 'obscene', 'model_response'],
                                'probas': [0, 0, 0.006988049950450659]},
                'text_sentiment': {'classes': ['negative', 'positive', 'neutral'],
                                   'probas': [0.000301319727441296, 0.06846659630537033, 0.9312320351600647]},
                'asr_sentiment': {'classes': [], 'probas': []}
            },
            'selected_item': {},
            'new_session': False,
            'strategies': {'last_call': None},
            'character': {'id': 'sber', 'name': 'Сбер', 'gender': 'male', 'appeal': 'official'},
            'meta': {'current_app': {}, 'time': {}},
            'asr': {}
        },
        'uuid': {
            'userId': '803de100-edd6-41d8-6d77',
            'sub': 'wXjow3XPPKlnD6lq4EnEIIimjqieg4hilLy8PZvF++',
            'userChannel': 'B2C'
        }
    }
    return result


def test_salut_request(salut_req):
    adapter = SalutAdapter()
    ctx = adapter.make_context(salut_req)
    assert ctx.message_text == 'настоящий'


def test_salut_response(salut_req):
    adapter = SalutAdapter()
    response = Response(
        text='kek',
        voice='brekekek',
        suggests=['yes', 'no'],
        commands=[COMMANDS.EXIT],
        links=[{'title': 'click me', 'url': 'www.com'}]
    )
    result = adapter.make_response(response=response, original_message=salut_req)
    assert len(result['uuid']) == 3
    assert result['payload']['pronounceText'] == 'brekekek'
    assert result['payload']['items'] == [{'bubble': {'text': 'kek'}}]
    assert len(result['payload']['suggestions']['buttons']) == 3
    assert 'device' in result['payload']
    assert result['payload']['finished'] is True



import dialogic


def test_create_flask_server():
    server = dialogic.server.flask_server.FlaskServer(
        connector=dialogic.dialog_connector.DialogConnector(
            dialog_manager=dialogic.dialog_manager.BaseDialogManager()
        )
    )



import pytest
import dialogic

from dialogic.dialog_manager.automaton import AutomatonDialogManager
from dialogic.testing.testing_utils import make_context


def test_empty_fsa():
    # FSA cannot be empty, because initial state must exist
    with pytest.raises(ValueError):
        fsa = AutomatonDialogManager(config={}, matcher='exact')


def test_minimal_fsa():
    cfg = {
        'states': {
            'first': {'a': 'hello'}
        }
    }
    fsa = AutomatonDialogManager(config=cfg, matcher='exact')
    assert len(fsa.states) == 2  # 'universal' state is always added but it never reached

    ctx = make_context(text='hi', new_session=True)
    resp = fsa.try_to_respond(ctx)
    assert resp.text == 'hello'


@pytest.fixture
def example_fsa():
    cfg = {
        'states': {
            'first': {'a': 'hello', 'next': [{'intent': 'time', 'label': 'second'}]},
            'second': {'a': '8 pm'},
            'third': {'q': ['help', 'what can you do'], 'a': 'I am always here to help'},
            'fourth': {'a': 'hello again', 'next': [{'intent': 'time', 'label': 'second'}]},
            'fifth': {'q': ['thanks'], 'a': 'thank you', 'restore_prev_state': True},
            'intro': {'q': ['let\'s introduce ourselves'], 'a': 'I am Max. And you?', 'default_next': 'my_name'},
            'my_name': {'a': 'Nice to meet you'},
        },
        'intents': {
            'time': {'regex': '.*time.*'}
        },
        'options': {
            'state_on_new_session': 'fourth',
        },
    }
    fsa = AutomatonDialogManager(config=cfg, matcher='exact')
    return fsa


def test_basic_transition(example_fsa):
    fsa: AutomatonDialogManager = example_fsa

    # initialize
    ctx0 = make_context(text='hi', new_session=True)
    resp0 = fsa.try_to_respond(ctx0)
    assert resp0.text == 'hello'

    # successful transition
    ctx1 = make_context(text='can you tell me the time please', prev_response=resp0)
    resp1 = fsa.try_to_respond(ctx1)
    assert resp1.text == '8 pm'

    # the same transition from another state is not allowed
    ctx2 = make_context(text='can you tell me the time please', prev_response=resp1)
    resp2 = fsa.try_to_respond(ctx2)
    assert not resp2

    # failed transition: text was not matched
    ctx1 = make_context(text='you will not understand me', prev_response=resp0)
    resp1 = fsa.try_to_respond(ctx1)
    assert not resp1

    # transition from the universal state
    ctx1 = make_context(text='help', prev_response=resp0)
    resp1 = fsa.try_to_respond(ctx1)
    assert resp1.text == 'I am always here to help'

    # new session
    ctx2 = make_context(new_session=True, prev_response=resp1)
    resp2 = fsa.try_to_respond(ctx2)
    assert resp2.text == 'hello again'

    # after transient state, context is restored and previous transition is possible
    ctx3 = make_context(prev_response=resp2, text='thanks')
    resp3 = fsa.try_to_respond(ctx3)
    assert resp3.text == 'thank you'
    ctx4 = make_context(prev_response=resp3, text='tell me time now')
    resp4 = fsa.try_to_respond(ctx4)
    assert resp4.text == '8 pm'


def test_default_transition(example_fsa):
    fsa: AutomatonDialogManager = example_fsa

    ctx0 = make_context(text='hi', new_session=True)
    resp0 = fsa.try_to_respond(ctx0)

    ctx1 = make_context('let\'s introduce ourselves', prev_response=resp0)
    resp1 = fsa.try_to_respond(ctx1)

    ctx2 = make_context('Stasy', prev_response=resp1)
    resp2 = fsa.try_to_respond(ctx2)
    assert resp2.text == 'Nice to meet you'



from typing import Dict, Optional

from ..adapters.base import BaseAdapter, Context, Response, logger
from ..dialog.names import SOURCES, REQUEST_TYPES, COMMANDS
from ..interfaces.yandex import YandexRequest, YandexResponse
from ..utils.text import encode_uri
from ..utils.content_manager import YandexImageAPI


class AliceAdapter(BaseAdapter):
    SOURCE = SOURCES.ALICE

    def __init__(self, native_state=False, image_manager: Optional[YandexImageAPI] = None, **kwargs):
        super(AliceAdapter, self).__init__(**kwargs)
        self.native_state = native_state
        self.image_manager: Optional[YandexImageAPI] = image_manager

    def make_context(self, message: Dict, **kwargs) -> Context:
        metadata = {}

        if set(message.keys()) == {'body'}:
            message = message['body']
        try:
            sess = message['session']
        except KeyError:
            raise KeyError(f'The key "session" not found in message among keys {list(message.keys())}.')
        if sess.get('user', {}).get('user_id'):
            # the new user_id, which is persistent across applications
            user_id = self.SOURCE + '_auth__' + sess['user']['user_id']
        else:
            # the old user id, that changes across applications
            user_id = self.SOURCE + '__' + sess['user_id']
        try:
            message_text = message['request'].get('command', '')
        except KeyError:
            raise KeyError(f'The key "request" not found in message among keys {list(message.keys())}.')
        metadata['new_session'] = message.get('session', {}).get('new', False)

        ctx = Context(
            user_object=None,
            message_text=message_text,
            metadata=metadata,
            user_id=user_id,
            session_id=sess.get('session_id'),
            source=self.SOURCE,
            raw_message=message,
        )

        ctx.request_type = message['request'].get('type', REQUEST_TYPES.SIMPLE_UTTERANCE)
        ctx.payload = message['request'].get('payload', {})
        try:
            ctx.yandex = YandexRequest.from_dict(message)
        except Exception as e:
            logger.warning('Could not deserialize Yandex request: got exception "{}".'.format(e))

        return ctx

    def make_response(self, response: Response, original_message=None, **kwargs):

        directives = {}
        if response.commands:
            for command in response.commands:
                if command == COMMANDS.REQUEST_GEOLOCATION:
                    directives[COMMANDS.REQUEST_GEOLOCATION] = {}

        result = {
            "version": original_message['version'],
            "response": {
                "end_session": response.has_exit_command,
                "text": response.text
            }
        }
        if self.native_state and response.updated_user_object:
            if self.native_state == 'session':
                result['session_state'] = response.updated_user_object
            elif self.native_state == 'application':
                result['application_state'] = response.updated_user_object
            elif self.native_state == 'user':
                if original_message.get('session') and 'user' not in original_message['session']:
                    result['application_state'] = response.updated_user_object
                result['user_state_update'] = response.updated_user_object
            else:
                if 'session' in response.updated_user_object:
                    result['session_state'] = response.updated_user_object['session']
                if 'application' in response.updated_user_object:
                    result['application_state'] = response.updated_user_object['application']
                if 'user' in response.updated_user_object:
                    result['user_state_update'] = response.updated_user_object['user']
        if response.raw_response is not None:
            if isinstance(response.raw_response, YandexResponse):
                result = response.raw_response.to_dict()
            else:
                result['response'] = response.raw_response
            return result
        if response.voice is not None and response.voice != response.text:
            result['response']['tts'] = response.voice.replace('\n', ' ')
        buttons = response.links or []
        for button in buttons:
            # avoid cyrillic characters in urls - they are not supported by Alice
            if 'url' in button:
                button['url'] = encode_uri(button['url'])
        if response.suggests:
            buttons = buttons + [{'title': suggest} for suggest in response.suggests]
        for button in buttons:
            if not isinstance(button.get('hide'), bool):
                button['hide'] = True
        result['response']['buttons'] = buttons
        if response.image_id:
            result['response']['card'] = {
                'type': 'BigImage',
                'image_id': response.image_id,
                'description': response.text
            }
        elif response.image_url and self.image_manager:
            image_id = self.image_manager.get_image_id_by_url(response.image_url)
            if image_id:
                result['response']['card'] = {
                    'type': 'BigImage',
                    'image_id': image_id,
                    'description': response.text
                }
        if response.gallery is not None:
            result['response']['card'] = response.gallery.to_dict()
        if response.image is not None:
            result['response']['card'] = response.image.to_dict()
        if response.show_item_meta is not None:
            result['response']['show_item_meta'] = response.show_item_meta
        if directives:
            result['response']['directives'] = directives
        return result

    def uses_native_state(self, context: Context) -> bool:
        """ Whether dialog state can be extracted directly from a context"""
        return bool(self.native_state)

    def get_native_state(self, context: Context) -> Optional[Dict]:
        """ Return native dialog state if it is possible"""
        if not self.native_state:
            return
        message = context.raw_message
        state = message.get('state', {})

        if self.native_state == 'session':
            user_object = state.get('session')
        elif self.native_state == 'user':
            user_object = state.get('user')
            # for unauthorized users, use application state instead
            if message.get('session') and 'user' not in message['session']:
                user_object = state.get('application')
        elif self.native_state == 'application':
            user_object = state.get('application')
        else:
            user_object = state
        return user_object



import dialogic

from dialogic.dialog.phrase import Phrase


def test_phrase_from_string():
    p1 = Phrase.from_object('hello')
    assert p1.texts == ['hello']
    resp = p1.render()
    assert resp.text == resp.voice == 'hello'

    p2 = Phrase.from_object('<text>Ciao</text><voice>Чао</voice>')
    resp = p2.render()
    assert resp.text == 'Ciao'
    assert resp.voice == 'Чао'


def test_phrase_from_dict():
    p1 = Phrase.from_object({
        'name': 'greeting',
        'text': 'hello',
        'exit': True,
        'suggests': ['Ciao', 'Bye'],
    })
    resp = p1.render(additional_suggests=['Auf Wiedersehen'])
    assert resp.commands == [dialogic.dialog.names.COMMANDS.EXIT]
    assert resp.suggests == ['Ciao', 'Bye', 'Auf Wiedersehen']


def test_random_phrase():
    p1 = Phrase(name='hi', text=['hello', 'hi'])

    assert len({p1.render(seed=1).text for i in range(100)}) == 1
    assert len({p1.render().text for i in range(100)}) == 2



"""
This module has been copied from https://github.com/gstaff/flask-ngrok with a few small modification
"""

import atexit
import json
import os
import platform
import shutil
import subprocess
import tempfile
import time
import zipfile
from pathlib import Path
from threading import Timer

import requests


def _get_command():
    system = platform.system()
    if system == "Darwin":
        command = "ngrok"
    elif system == "Windows":
        command = "ngrok.exe"
    elif system == "Linux":
        command = "ngrok"
    else:
        raise Exception("{system} is not supported".format(system=system))
    return command


def run_ngrok(port, wait=5, retries=10):
    command = _get_command()
    ngrok_path = str(Path(tempfile.gettempdir(), "ngrok"))
    _download_ngrok(ngrok_path)
    executable = str(Path(ngrok_path, command))
    os.chmod(executable, 0o777)
    ngrok = subprocess.Popen([executable, 'http', str(port)])
    atexit.register(ngrok.terminate)
    localhost_url = "http://localhost:4040/api/tunnels"  # Url with tunnel details
    time.sleep(1)
    tunnel_url = None

    for i in range(retries):
        try:
            tunnel_url = requests.get(localhost_url).text  # Get the tunnel information
            break
        except requests.exceptions.ConnectionError as e:
            print('Will retry connecting to ngrok api, got error {}'.format(e))
            time.sleep(wait)
    if tunnel_url is None:
        raise ValueError('Could not connect to ngrok api, exiting')
    j = json.loads(tunnel_url)

    tunnel_url = j['tunnels'][0]['public_url']  # Do the parsing of the get
    tunnel_url = tunnel_url.replace("https", "http")
    return tunnel_url


def _download_ngrok(ngrok_path):
    if Path(ngrok_path).exists():
        return
    system = platform.system()
    if system == "Darwin":
        url = "https://bin.equinox.io/c/4VmDzA7iaHb/ngrok-stable-darwin-amd64.zip"
    elif system == "Windows":
        url = "https://bin.equinox.io/c/4VmDzA7iaHb/ngrok-stable-windows-amd64.zip"
    elif system == "Linux":
        url = "https://bin.equinox.io/c/4VmDzA7iaHb/ngrok-stable-linux-amd64.zip"
    else:
        raise Exception("{} is not supported".format(system))
    download_path = _download_file(url)
    with zipfile.ZipFile(download_path, "r") as zip_ref:
        zip_ref.extractall(ngrok_path)


def _download_file(url):
    local_filename = url.split('/')[-1]
    r = requests.get(url, stream=True)
    download_path = str(Path(tempfile.gettempdir(), local_filename))
    with open(download_path, 'wb') as f:
        shutil.copyfileobj(r.raw, f)
    return download_path


def start_ngrok(port):
    ngrok_address = run_ngrok(port)
    print(" * Running on {}".format(ngrok_address))
    print(" * Traffic stats available on http://127.0.0.1:4040")


def run_with_ngrok(app):
    """
    The provided Flask app will be securely exposed to the public internet via ngrok when run,
    and the its ngrok address will be printed to stdout
    :param app: a Flask application object
    :return: None
    """
    old_run = app.run

    def new_run(*args, **kwargs):
        port = kwargs.get('port', 5000)
        thread = Timer(1, start_ngrok, args=(port,))
        thread.setDaemon(True)
        thread.start()
        old_run(*args, **kwargs)
    app.run = new_run



import inspect
from collections.abc import Mapping


def list_converter(cls):
    def converter(item):
        return [cls.from_dict(x) for x in (item or [])]
    return converter


def dict_converter(cls):
    def converter(item):
        return {k: cls.from_dict(v) for k, v in (item or {}).items()}
    return converter


def try_serialize(item):
    if hasattr(item, 'to_dict'):
        return item.to_dict()
    return item


class Serializeable:
    """ This mixin is used for easy conversion of structured objects from and to json """

    @classmethod
    def from_dict(cls, data, from_none=False):
        if data is None:
            if from_none:
                data = {}
            else:
                return None
        if isinstance(data, cls):
            return data
        # assume all the heavy lifting is done by converters in the cls.__init__
        return cls(**data)

    def to_dict(self):
        result = {}
        for field_name, field in self.__dict__.items():
            if isinstance(field, list):
                field = [try_serialize(item) for item in field]
            elif isinstance(field, dict):
                field = {k: try_serialize(v) for k, v in field.items()}
            else:
                field = try_serialize(field)
            result[field_name] = field
        return result


class FreeSerializeable(Serializeable):
    """ A serializeable object that can accept and preserve arbitrary extra keys. """
    @classmethod
    def from_dict(cls, data, from_none=False):
        if isinstance(data, Mapping) and not isinstance(data, cls):
            arg_names = set(inspect.signature(cls.__init__).parameters)
            args = {k: v for k, v in data.items() if k in arg_names}
            other = {k: v for k, v in data.items() if k not in arg_names}
        else:
            args = data
            other = None
        result = super(FreeSerializeable, cls).from_dict(args, from_none=from_none)
        if result is not None:
            result._other = other
        return result

    def to_dict(self):
        result = super(FreeSerializeable, self).to_dict()
        if hasattr(self, '_other') and self._other:
            result.update(self._other)
        if '_other' in result:
            del result['_other']
        return result



from typing import Dict, Optional

from ..dialog.serialized_message import SerializedMessage
from ..dialog.names import SOURCES, REQUEST_TYPES
from ..adapters.base import BaseAdapter, Context, Response, logger


class VkAdapter(BaseAdapter):
    SOURCE = SOURCES.VK

    def __init__(
            self,
            suggest_cols=1, suggest_screen=32, suggest_margin=1, suggest_max_len=40,
            **kwargs
    ):
        super(VkAdapter, self).__init__(**kwargs)
        self.suggest_cols = suggest_cols
        self.suggest_screen = suggest_screen
        self.suggest_margin = suggest_margin
        self.suggest_max_len = suggest_max_len

    def make_context(self, message, **kwargs) -> Context:
        uid = self.SOURCE + '__' + str(message.user_id)
        ctx = Context(
            user_object=None,
            message_text=message.text,
            metadata={},
            user_id=uid,
            session_id=uid,
            source=self.SOURCE,
            raw_message=message,
        )
        return ctx

    def make_response(self, response: Response, original_message=None, **kwargs):
        # todo: instead of a dict, use a class object as a response
        # todo: add multimedia, etc.
        result = {
            'text': response.text,
        }
        if response.suggests or response.links:
            buttons = []
            for i, link in enumerate(response.links):
                buttons.append({'action': {'type': 'open_link', 'label': link['title'], 'link': link['url']}})
            for i, suggest in enumerate(response.suggests):
                buttons.append({'action': {'type': 'text', 'label': suggest}})

            rows = []
            row_width = 0
            for i, button in enumerate(buttons):
                if self.suggest_cols == 'auto':
                    extra_width = len(button['action']['label']) + self.suggest_margin * 2
                    if len(rows) == 0 or row_width > 0 and row_width + extra_width > self.suggest_screen:
                        rows.append([])
                        row_width = extra_width
                    else:
                        row_width += extra_width
                else:
                    if i % self.suggest_cols == 0:
                        rows.append([])
                rows[-1].append(button)

            for row in rows:
                for button in row:
                    label = button['action']['label']
                    if len(label) > self.suggest_max_len:
                        button['action']['label'] = label[:(self.suggest_max_len - 3)] + '...'

            result['keyboard'] = {
                'one_time': True,
                'buttons': rows,
            }
        return result

    def serialize_context(self, context: Context, data=None, **kwargs) -> Optional[SerializedMessage]:
        serializable_message = {'message': context.raw_message.to_json()}
        return super(VkAdapter, self).serialize_context(context=context, data=serializable_message, **kwargs)



from . import automaton, base, faq, form_filling
from .base import (
    Context, Response, BaseDialogManager, CascadableDialogManager, CascadeDialogManager, GreetAndHelpDialogManager,
    COMMANDS
)
from .faq import FAQDialogManager
from .form_filling import FormFillingDialogManager
from .automaton import AutomatonDialogManager
from .turning import TurnDialogManager



import copy
from typing import Dict, Optional

import telebot
from ..dialog.serialized_message import SerializedMessage
from ..dialog.names import SOURCES
from dialogic.adapters.base import BaseAdapter, Context, Response, logger
from telebot.types import Message


class TelegramAdapter(BaseAdapter):
    SOURCE = SOURCES.TELEGRAM

    def __init__(self, suggest_cols=1, **kwargs):
        super(TelegramAdapter, self).__init__(**kwargs)
        self.suggest_cols = suggest_cols

    def make_context(self, message: Message, **kwargs) -> Context:
        uid = self.SOURCE + '__' + str(message.from_user.id)
        ctx = Context(
            user_object=None,
            message_text=message.text,
            metadata={},
            user_id=uid,
            session_id=uid,
            source=self.SOURCE,
            raw_message=message,
        )
        return ctx

    def make_response(self, response: Response, original_message=None, **kwargs):
        if response.raw_response is not None:
            return response.raw_response
        result = {
            'text': response.text
        }
        if response.links is not None:
            result['parse_mode'] = 'html'
            for link in response.links:
                result['text'] += '\n<a href="{}">{}</a>'.format(link['url'], link['title'])
        if response.suggests:
            # todo: do smarter row width calculation
            row_width = min(self.suggest_cols or 1, len(response.suggests))
            result['reply_markup'] = telebot.types.ReplyKeyboardMarkup(row_width=row_width)
            result['reply_markup'].add(*[telebot.types.KeyboardButton(t) for t in response.suggests])
        else:
            result['reply_markup'] = telebot.types.ReplyKeyboardRemove(selective=False)
        if response.image_url:
            if 'multimedia' not in result:
                result['multimedia'] = []
            media_type = 'document' if response.image_url.endswith('.gif') else 'photo'
            result['multimedia'].append({'type': media_type, 'content': response.image_url})
        if response.sound_url:
            if 'multimedia' not in result:
                result['multimedia'] = []
            result['multimedia'].append({'type': 'audio', 'content': response.sound_url})
        result['disable_web_page_preview'] = True
        return result

    def serialize_context(self, context: Context, data=None, **kwargs) -> Optional[SerializedMessage]:
        message = context.raw_message
        if message.reply_to_message is not None:
            kwargs['reply_to_id'] = message.reply_to_message.message_id
        kwargs['message_id'] = message.message_id
        kwargs['username'] = message.chat.username
        serializable_message = {'message': str(message)}
        return super(TelegramAdapter, self).serialize_context(context=context, data=serializable_message, **kwargs)

    def serialize_response(self, data, context: Context, response: Response, **kwargs) -> Optional[SerializedMessage]:
        data = copy.deepcopy(data)
        message = context.raw_message
        kwargs['reply_to_id'] = message.message_id
        kwargs['username'] = message.chat.username
        # todo: maybe somehow get message_id for output messages
        if 'reply_markup' in data:
            data['reply_markup'] = data['reply_markup'].to_json()
        return super(TelegramAdapter, self).serialize_response(
            data=data,
            context=context,
            response=response,
            kwargs=kwargs
        )




class SOURCES:
    ALICE = 'alice'
    FACEBOOK = 'facebook'
    SALUT = 'salut'
    TELEGRAM = 'telegram'
    TEXT = 'text'
    VK = 'vk'
    unknown_source_error_message = 'Source must be on of {"alice", "facebook", "telegram", "text", "vk"}'


class COMMANDS:
    EXIT = 'exit'
    REQUEST_GEOLOCATION = 'request_geolocation'


class REQUEST_TYPES:
    SIMPLE_UTTERANCE = 'SimpleUtterance'
    PUSH = 'push'
    BUTTON_PRESSED = 'ButtonPressed'
    SHOW_PULL = 'Show.Pull'
    GEOLOCATION_ALLOWED = 'Geolocation.Allowed'
    GEOLOCATION_REJECTED = 'Geolocation.Rejected'



"""
This module allows creating reusable parts of regular expressions.
It might be useful for matching intents or extracting slots.
The resulting language is not as powerful as grammars, but close to it.

The example below consists of two data files and a code file:
The file `expressions.yaml`:
```
NUMBER: '[0-9]+'
_NUMBER_GROUP: '(number )?(?P<id>{{NUMBER}})'
```
The file `intents.yaml`:
```
choose:
    regex: '((choose|take|set) )?{{_NUMBER_GROUP}}'
```
The file `code.py`:
```
intents = load_intents_with_replacement('intents.yaml', 'expressions.yaml')
matcher = make_matcher_with_regex(TFIDFMatcher(), intents=intents)
print(matcher.match('take number 10'))
"""

import re
import yaml


def _patch(text, expressions):
    def sub_maker(match):
        return '({})'.format(expressions[match.group(1)])
    return re.sub('{{\\s*([a-zA-Z_]+)\\s*}}', sub_maker, text)


def load_expressions(filename):
    with open(filename, encoding='utf-8') as f:
        expressions = yaml.safe_load(f)
    for k in list(expressions.keys()):
        expressions[k] = _patch(expressions[k], expressions)
    return expressions


def load_intents_with_replacement(intents_fn, expressions_fn):
    expressions = load_expressions(expressions_fn)
    with open(intents_fn, encoding='utf-8') as f:
        intents = yaml.safe_load(f)
    for v in intents.values():
        if 'regexp' in v:
            if isinstance(v['regexp'], list):
                v['regexp'] = [_patch(e, expressions) for e in v['regexp']]
            elif isinstance(v['regexp'], str):
                v['regexp'] = _patch(v['regexp'], expressions)
    return intents



import json
import sys

from collections.abc import Iterable, Mapping
from collections import OrderedDict


def make_button(button=None, text=None, url=None, payload=None):
    if isinstance(button, GalleryButton):
        return button
    elif isinstance(button, Mapping):
        return GalleryButton(**button)
    elif text is not None or url is not None or payload is not None:
        return GalleryButton(text=text, url=url, payload=payload)
    else:
        return None


class Button:
    def __init__(self, title=None, payload=None, url=None, hide=True):
        # todo: write getters and setters with validation
        self.title = title
        if self.title is None:
            raise ValueError('Button title cannot be empty')
        else:
            assert len(self.title) <= 64, 'Button title cannot be longer than 64 symbols'
        self.payload = payload
        if self.payload is not None:
            assert sys.getsizeof(json.dumps(self.payload)) <= 4096, 'Size of button payload cannot exceed 4096 bytes'
        self.url = url
        if self.url is not None:
            assert sys.getsizeof(self.url) <= 1024, 'Size of button URL cannot exceed 1024 bytes'
        self.hide = bool(hide)

    def to_dict(self):
        result = OrderedDict()
        if self.title is not None:
            result['title'] = self.title
        if self.payload is not None:
            result['payload'] = self.payload
        if self.url is not None:
            result['url'] = self.url
        result['hide'] = self.hide
        return result


class BigImage:
    def __init__(self, image_id=None, title=None, description=None,
                 button=None, button_text=None, button_url=None, button_payload=None):
        self.image_id = image_id
        self.title = title
        self.description = description
        if self.description is not None:
            assert len(self.description) <= 256, 'Big Image description cannot be longer than 256 symbols'
        self.button = make_button(button, button_text, button_url, button_payload)

    def to_dict(self):
        result = OrderedDict()
        result['type'] = 'BigImage'
        result['image_id'] = self.image_id
        if self.title is not None:
            result['title'] = self.title
        if self.description is not None:
            result['description'] = self.description
        if self.button is not None:
            result['button'] = self.button.to_dict()
        return result


class GalleryButton:
    def __init__(self, text=None, payload=None, url=None):
        # todo: write getters and setters with validation
        self.text = text
        if self.text is not None:
            assert len(self.text) <= 64, 'Button title cannot be longer than 64 symbols'
        self.payload = payload
        if self.payload is not None:
            assert sys.getsizeof(json.dumps(self.payload)) <= 4096, 'Size of button payload cannot exceed 4096 bytes'
        self.url = url
        if self.url is not None:
            assert sys.getsizeof(self.url) <= 1024, 'Size of button URL cannot exceed 1024 bytes'

    def to_dict(self):
        result = OrderedDict()
        if self.text is not None:
            result['text'] = self.text
        if self.payload is not None:
            result['payload'] = self.payload
        if self.url is not None:
            result['url'] = self.url
        return result


class GalleryItem:
    def __init__(self, image_id=None, title=None, description=None,
                 button=None, button_text=None, button_url=None, button_payload=None):
        self.image_id = image_id
        self.title = title
        self.description = description
        if self.description is not None:
            assert len(self.description) <= 256, 'Gallery item description cannot be longer than 256 symbols'
        self.button = make_button(button, button_text, button_url, button_payload)

    def to_dict(self):
        result = OrderedDict()
        if self.image_id is not None:
            result['image_id'] = self.image_id
        if self.title is not None:
            result['title'] = self.title
        if self.description is not None:
            result['description'] = self.description
        if self.button is not None:
            result['button'] = self.button.to_dict()
        return result


class GalleryFooter:
    def __init__(self, text=None,
                 button=None, button_text=None, button_url=None, button_payload=None):
        self.text = text
        if self.text is not None:
            assert len(self.text) <= 64, 'Gallery footer text cannot be longer than 64 symbols'
        self.button = make_button(button, button_text, button_url, button_payload)

    def to_dict(self):
        result = OrderedDict()
        if self.text is not None:
            result['text'] = self.text
        if self.button is not None:
            result['button'] = self.button.to_dict()
        return result


class Gallery:
    def __init__(self, title=None, items=None, footer=None):
        self.title = title
        if self.title is not None:
            assert len(self.title) <= 64, 'Gallery header text cannot be longer than 64 symbols'
        if not isinstance(items, Iterable):
            raise ValueError('Gallery items should be an Iterable, got {}'.format(type(items)))
        if items is None:
            items = []
        if len(items) < 1 or len(items) > 5:
            raise ValueError('Gallery should contain 1-5 items, got {}'.format(len(items)))
        self.items = items
        if footer is not None:
            assert isinstance(footer, GalleryFooter)
            # todo: make footer create-able from config
        self.footer = footer

    def to_dict(self):
        result = OrderedDict()
        result['type'] = 'ItemsList'
        if self.title is not None:
            result['header'] = {'text': self.title}
        result['items'] = [item.to_dict() for item in self.items]
        if self.footer is not None:
            result['footer'] = self.footer.to_dict()
        return result



from dialogic.dialog_manager import Context


def make_context(text='', prev_response=None, new_session=False):
    if prev_response is not None:
        user_object = prev_response.updated_user_object
    else:
        user_object = {}
    if new_session:
        metadata = {'new_session': True}
    else:
        metadata = {}
    return Context(user_object=user_object, metadata=metadata, message_text=text)



import logging
import time
import yaml
from typing import Union, Type, Tuple, Dict

from ..interfaces.yandex import extract_yandex_forms
from ..nlu.regex_utils import match_forms
from dialogic.cascade import DialogTurn
from dialogic.dialog import Context, Response
from dialogic.dialog_manager import CascadableDialogManager
from dialogic.nlu.basic_nlu import fast_normalize
from dialogic.nlu.matchers import TFIDFMatcher, TextNormalization, make_matcher_with_regex

logger = logging.getLogger(__name__)


class TurnDialogManager(CascadableDialogManager):
    def __init__(
            self,
            cascade,
            intents_file=None,
            turn_cls: Type[DialogTurn] = DialogTurn,
            **kwargs
    ):
        super(TurnDialogManager, self).__init__(**kwargs)

        self.cascade = cascade
        self.turn_cls: Type[DialogTurn] = turn_cls

        self.intents_file = intents_file
        self.intents = {}
        self.intent_matcher = None
        if intents_file:
            self.load_intents(intents_file=intents_file)

    def load_intents(self, intents_file=None):
        with open(intents_file, 'r', encoding='utf-8') as f:
            self.intents = yaml.safe_load(f)

        self.intent_matcher = make_matcher_with_regex(
            base_matcher=TFIDFMatcher(text_normalization=TextNormalization.FAST_LEMMATIZE),
            intents=self.intents,
        )

    def try_to_respond(self, ctx: Context) -> Union[Response, None]:
        t = time.time()
        self.preprocess_context(ctx=ctx)
        text, intents, forms = self.nlu(ctx=ctx)
        turn = self.turn_cls(
            ctx=ctx,
            text=text,
            intents=intents,
            forms=forms,
            user_object=ctx.user_object or {},
        )
        self.preprocess_turn(turn=turn)
        handler_name = self.cascade(turn)
        logger.debug(f"Final handler: {handler_name}")
        response = turn.make_response()
        self.postprocess_response(response=response, turn=turn)
        logger.debug(f'DM response took {time.time() - t} seconds')
        return response

    def nlu(self, ctx: Context) -> Tuple[str, Dict[str, float], Dict[str, Dict]]:
        text = fast_normalize(ctx.message_text or '')
        if self.intent_matcher:
            intents = self.intent_matcher.aggregate_scores(text)
        else:
            intents = {}
        forms = match_forms(text=text, intents=self.intents or {})
        if ctx.yandex:
            ya_forms = extract_yandex_forms(ctx.yandex)
            forms.update(ya_forms)
            for intent_name in ya_forms:
                intents[intent_name] = 1
        return text, intents, forms

    def preprocess_context(self, ctx: Context):
        if not ctx.user_object:
            ctx.add_user_object({})
        if ctx.session_is_new():
            if 'stage' in ctx.user_object:
                del ctx.user_object['stage']
            ctx.user_object['sessions_count'] = ctx.user_object.get('sessions_count', 0) + 1

    def preprocess_turn(self, turn: DialogTurn):
        pass

    def postprocess_response(self, response: Response, turn):
        pass



import math
import textdistance
import re
import typing

from collections import Counter, defaultdict
from collections.abc import Callable, Iterable, Mapping
from itertools import chain
from types import ModuleType

from ..nlu import basic_nlu

from .regex_utils import regex

try:
    from pyemd import emd
    IMPORTED_EMD = True
except ImportError:
    emd = None
    IMPORTED_EMD = False

try:
    import numpy as np
    IMPORTED_NUMPY = True
except ImportError:
    np = None
    IMPORTED_NUMPY = False


EPSILON = 1e-10


class TextNormalization:
    FAST = 'fast'
    FAST_LEMMATIZE = 'fast_lemmatize'


class BaseMatcher:
    """ A base class for text classification with confidence """
    def __init__(self, threshold: float = 0.5, thresholds=None):
        """ Create a base matcher
        parameters:
        - threshold: the minimal allowed similarity between the text and the example for a successful match
        - thresholds: an optional dict of label-wise thresholds
        """
        self.threshold: float = threshold
        self.thresholds: dict = thresholds or {}

    def fit(self, texts, labels):
        raise NotImplementedError()

    def get_threshold(self, label):
        return self.thresholds.get(label, self.threshold)

    def match(self, text: str, use_threshold=True):
        """ Return the label which is most similar to the text and its score.
        If no example is similar enough, the winner label will be None.
        """
        scores, labels = self.get_scores(text)
        best_score = -math.inf
        winner_label = None
        for score, label in zip(scores, labels):
            if (score >= self.get_threshold(label) or not use_threshold) and score > best_score:
                best_score = score
                winner_label = label
        return winner_label, best_score

    def get_scores(self, text: str) -> typing.Tuple[typing.List[float], typing.List]:
        """ Return the list of matching scores and their corresponding labels """
        raise NotImplementedError()

    def aggregate_scores(self, text: str, use_threshold=True) -> Counter:
        """ Return a dict with the highest matching score for each label. """
        result = Counter()
        scores, labels = self.get_scores(text)
        for score, label in zip(scores, labels):
            if score >= self.get_threshold(label) or not use_threshold:
                result[label] = max(score, result.get(label, -math.inf))
        return result


class AggregationMatcher(BaseMatcher):
    def __init__(self, matchers, **kwargs):
        super(AggregationMatcher, self).__init__(**kwargs)
        assert len(matchers) > 0, 'list of matchers should be non-empty'
        self.matchers = matchers

    def fit(self, texts, labels):
        for m in self.matchers:
            m.fit(texts, labels)

    def get_scores(self, text: str) -> typing.Tuple[typing.List[float], typing.List]:
        scores = []
        labels = []
        for m in self.matchers:
            sc, lbl = m.get_scores(text)
            scores.extend(sc)
            labels.extend(lbl)
        return scores, labels

    def _apply_matchers(self, text, use_threshold=False):
        label2matchers2scores = defaultdict(lambda: defaultdict(lambda: -math.inf))
        for i, m in enumerate(self.matchers):
            scores, labels = m.get_scores(text)
            for label, score in zip(labels, scores):
                if score > label2matchers2scores[label][i]:
                    if use_threshold and score < self.get_threshold(label):
                        continue
                    label2matchers2scores[label][i] = score
        return label2matchers2scores


class MaxMatcher(AggregationMatcher):
    def get_scores(self, text):
        label2matchers2scores = self._apply_matchers(text)
        labels = []
        scores = []
        for label, ldict in label2matchers2scores.items():
            if not ldict:
                continue
            labels.append(label)
            scores.append(max(ldict.values()))
        return scores, labels


class WeightedAverageMatcher(AggregationMatcher):
    def __init__(self, matchers, weights=None, **kwargs):
        super(WeightedAverageMatcher, self).__init__(matchers, **kwargs)
        if weights is None:
            weights = [1.0] * len(self.matchers)
        else:
            assert len(weights) == len(matchers), 'matchers and weights should have the same size'
        total_weight = float(sum(weights))
        self.weights = [w / total_weight for w in weights]

    def get_scores(self, text):
        label2matchers2scores = self._apply_matchers(text)
        labels = []
        scores = []
        for label, ldict in label2matchers2scores.items():
            score = sum(w * ldict[i] for i, w in enumerate(self.weights))
            labels.append(label)
            scores.append(score)
        return scores, labels


class ModelBasedMatcher(BaseMatcher):
    """ Classify text using an old good sklearn-style model """
    def __init__(self, model, **kwargs):
        super(ModelBasedMatcher, self).__init__(**kwargs)
        self.model = model

    def fit(self, texts, labels):
        self.model.fit(texts, labels)

    def get_scores(self, text):
        scores = self.model.predict_proba([text])[0]
        labels = self.model.classes_
        return scores, labels


class RegexMatcher(BaseMatcher):
    """ This matcher returns matching score 1,
    if the text matches one of the provided expressions for the label, and 0 otherwise.
    The parameter `add_end` forcibly wraps each expression between `^` and `$` symbols, disabling partial prefix match.
    """
    def __init__(self, *args, add_end=True, merge=True, engine='re', **kwargs):
        super(RegexMatcher, self).__init__(*args, **kwargs)
        self.add_end = add_end
        self.merge = merge
        self.expressions = []
        self.labels = []
        self.engine = engine

    @property
    def re(self):
        if self.engine == 'regex' and regex is not None:
            return regex
        elif isinstance(self.engine, ModuleType):
            return self.engine
        return re

    def fit(self, texts, labels):
        parts = defaultdict(list)
        for text, label in zip(texts, labels):
            parts[label].append(text)
        for label, expressions in parts.items():
            if self.merge:
                self.expressions.append(
                    self.re.compile('(?:{})'.format(
                        '|'.join([self._wrap(e) for e in expressions])
                    ))
                )
                self.labels.append(label)
            else:
                for e in expressions:
                    self.expressions.append(self.re.compile('(?:{})'.format(self._wrap(e))))
                    self.labels.append(label)

    def _wrap(self, text):
        if self.add_end:
            return '^{}$'.format(text)
        return text

    def get_scores(self, text):
        scores = []
        labels = []
        for label, expression in zip(self.labels, self.expressions):
            labels.append(label)
            scores.append(float(bool(self.re.match(expression, text))))
        return scores, labels


class PairwiseMatcher(BaseMatcher):
    """
    Classify text using 1-nearest neighbor by some similarity metric.
    This is an abstract class; its descendants should implement the specific metric to compare preprocessed texts.

    Parameters
    ----------
    text_normalization: string or callable
        Describes how to preprocess the texts before matching.
        Supported string values: 'fast' to lowercase and remove unusual characters; 'fast_lemmatize' to additionally
        lemmatize words (russian only). Callable values should accept and return strings.
    stopwords: iterable or mapping
        Lists the words that should be discarded (if it is a list) or paid less attention
        (if it is a dict with values in (0, 1) during matching. It may not be supported by all descendant matchers.
    kwargs:
        Passed to the parent constructor (BaseMatcher)
    """
    def __init__(self, text_normalization=TextNormalization.FAST, stopwords=None, **kwargs):
        super(PairwiseMatcher, self).__init__(**kwargs)
        self.text_normalization = text_normalization
        self._texts = []
        self._labels = []

        if stopwords is None:
            stopwords = {}
        if isinstance(stopwords, Iterable) and not isinstance(stopwords, Mapping):
            stopwords = {w: 0 for w in stopwords}
        # todo: delay stopwords preprocessing until the descendant class has been initialized
        stopwords = {PairwiseMatcher.preprocess(self, k): v for k, v in stopwords.items()}
        self.stopwords = stopwords

    def preprocess(self, text):
        if self.text_normalization == TextNormalization.FAST:
            text = basic_nlu.fast_normalize(text, lemmatize=False)
        if self.text_normalization == TextNormalization.FAST_LEMMATIZE:
            text = basic_nlu.fast_normalize(text, lemmatize=True)
        elif isinstance(self.text_normalization, Callable):
            text = self.text_normalization(text)
        return text

    def compare(self, one, another):
        raise NotImplementedError()

    def fit(self, texts, labels):
        self._texts = [self.preprocess(text) for text in texts]
        self._labels = labels
        return self

    def get_scores(self, text):
        processed = self.preprocess(text)
        return [self.compare(processed, t) for t in self._texts], self._labels


class ExactMatcher(PairwiseMatcher):
    def compare(self, one, another):
        return float(one == another)


class TextDistanceMatcher(PairwiseMatcher):
    def __init__(self, by_words=True, metric='cosine', **kwargs):
        super(TextDistanceMatcher, self).__init__(**kwargs)
        self.by_words = by_words
        self.metric = metric
        self.fun = getattr(textdistance, metric).normalized_similarity

    def preprocess(self, text):
        text = super(TextDistanceMatcher, self).preprocess(text)
        if self.by_words:
            return text.split()
        return text

    def compare(self, one, another):
        return self.fun(one, another)


class LevenshteinMatcher(TextDistanceMatcher):
    def __init__(self, **kwargs):
        kwargs['by_words'] = False
        kwargs['metric'] = 'levenshtein'
        super(LevenshteinMatcher, self).__init__(**kwargs)


class JaccardMatcher(PairwiseMatcher):
    def preprocess(self, text):
        text = super(JaccardMatcher, self).preprocess(text)
        return set(text.split())

    def compare(self, one, another):
        intersection = len(one.intersection(another))
        union = len(one.union(another))
        if intersection:
            return intersection / union
        return 0.0


class TFIDFMatcher(PairwiseMatcher):
    def __init__(self, smooth=2.0, ngram=1, **kwargs):
        super(TFIDFMatcher, self).__init__(**kwargs)
        self.smooth = smooth
        self.ngram = ngram
        self.vocab = Counter()

    def fit(self, texts, labels):
        self.vocab = Counter(w for t in texts for w in self._tokenize(super(TFIDFMatcher, self).preprocess(t)))
        return super(TFIDFMatcher, self).fit(texts, labels)

    def preprocess(self, text):
        text = super(TFIDFMatcher, self).preprocess(text)
        tf = Counter(self._tokenize(text))
        return {
            w: tf / math.log(self.smooth + self.vocab[w]) * self.stopwords.get(w, 1)
            for w, tf in tf.items()
        }

    def compare(self, one, another):
        dot = self._dot(one, another)
        if abs(dot) < 1e-6:
            return 0.0
        return dot / math.sqrt(self._norm(one) * self._norm(another))

    def _tokenize(self, text):
        words = text.split()
        if self.ngram == 1:
            return words
        words = ['BOS'] + words + ['EOS']
        return words + ['_'.join(words[i:(i + self.ngram)]) for i in range(len(words) - self.ngram + 1)]

    @staticmethod
    def _dot(one, another):
        return sum(v * another.get(k, 0) for k, v in one.items())

    def _norm(self, one):
        return self._dot(one, one)


class W2VMatcher(PairwiseMatcher):
    """ Compare texts by cosine similarity of their mean word vectors """
    def __init__(self, w2v, normalize_word_vec=True, **kwargs):
        super(W2VMatcher, self).__init__(**kwargs)
        self.w2v = w2v
        self.normalize_word_vec = normalize_word_vec

    def vec_from_word(self, word):
        vec = self.w2v[word]
        if self.normalize_word_vec:
            vec = vec / max(sum(vec**2), EPSILON) ** 0.5
        return vec

    def preprocess(self, text):
        text = super(W2VMatcher, self).preprocess(text)
        tokens = text.split()
        vecs = [self.vec_from_word(t) for t in tokens if t in self.w2v]
        if len(vecs) == 0:
            return None
        result = sum(vecs)
        result = result / max(sum(result**2), EPSILON) ** 0.5
        return result

    def compare(self, one, another):
        if one is None or another is None:
            return 0
        return sum(one * another)


class WMDDocument:
    def __init__(self, text, tokens, vecs, weights):
        self.text = text
        self.tokens = tokens
        self.vecs = vecs
        self.weights = weights


class WMDMatcher(PairwiseMatcher):
    """
    Compare texts by Word Mover Distance between them .

    When using this code, please consider citing the following papers:
        .. Ofir Pele and Michael Werman, "A linear time histogram metric for improved SIFT matching".
        .. Ofir Pele and Michael Werman, "Fast and robust earth mover's distances".
        .. Matt Kusner et al. "From Word Embeddings To Document Distances".
    """

    def __init__(self, w2v, normalize_word_vec=True, **kwargs):
        if not IMPORTED_NUMPY:
            raise ImportError('When using WMDMatcher, numpy should be installed')
        if not IMPORTED_EMD:
            raise ImportError('When using WMDMatcher, pyemd should be installed')
        super(WMDMatcher, self).__init__(**kwargs)
        self.w2v = w2v
        self.normalize_word_vec = normalize_word_vec

    def vec_from_word(self, word):
        vec = self.w2v[word]
        if self.normalize_word_vec:
            vec = vec / max(sum(vec ** 2), EPSILON) ** 0.5
        return vec

    def preprocess(self, text):
        preprocessed_text = super(WMDMatcher, self).preprocess(text)
        tokens = preprocessed_text.split()
        valid_tokens = [t for t in tokens if t in self.w2v]
        if len(valid_tokens) == 0:
            return None
        vecs = [self.vec_from_word(t) for t in valid_tokens]
        weights = []
        return WMDDocument(text, valid_tokens, vecs, weights)

    @staticmethod
    def text2bow(tokens, word2idx):
        bow = np.zeros(len(word2idx), dtype=np.double)
        n = len(tokens)
        for t in tokens:
            bow[word2idx[t]] += 1.0 / n
        return bow

    def compare(self, one, another):
        if one is None or another is None:
            return 0

        vocab = sorted(set(one.tokens).union(set(another.tokens)))
        word2idx = {word: i for i, word in enumerate(vocab)}
        word2vec = {}
        for w, v in chain(zip(one.tokens, one.vecs), zip(another.tokens, another.vecs)):
            if w not in word2vec:
                word2vec[w] = v

        distance_matrix = np.zeros((len(vocab), len(vocab)), dtype=np.double)
        for i, t1 in enumerate(vocab):
            for j, t2 in enumerate(vocab):
                if t1 not in one.tokens or t2 not in another.tokens:
                    continue
                # Compute Euclidean distance between word vectors.
                distance_matrix[i, j] = np.sqrt(np.sum((word2vec[t1] - word2vec[t2]) ** 2))

        d1 = self.text2bow(one.tokens, word2idx)
        d2 = self.text2bow(another.tokens, word2idx)

        wmd = emd(d1, d2, distance_matrix)
        # because we use unit vectors, this transformation mimics cosine distance
        similarity = 1 - wmd ** 2 / 2
        return similarity


def make_matcher_with_regex(base_matcher: BaseMatcher, intents, merge=True, re_matcher: RegexMatcher = None):
    """ Create a mix of the given matcher and a regex matcher """
    labels = []
    texts = []
    re_labels = []
    re_texts = []
    for intent_name, intent in intents.items():
        if 'regexp' in intent:
            exps = intent['regexp']
            if not isinstance(exps, list):
                exps = [exps]
            for exp in exps:
                re_labels.append(intent_name)
                re_texts.append(exp)
        if 'examples' in intent:
            for ex in intent['examples']:
                labels.append(intent_name)
                texts.append(ex)
    base_matcher.fit(texts, labels)
    if re_matcher is None:
        re_matcher = RegexMatcher(merge=merge)
    re_matcher.fit(re_texts, re_labels)
    return MaxMatcher([base_matcher, re_matcher])


_matchers = dict()


def register_matcher(name, matcher_maker):
    _matchers[name] = matcher_maker


def make_matcher(key, **kwargs):
    return _matchers[key](**kwargs)


register_matcher('exact', lambda **kwargs: ExactMatcher(**kwargs))
register_matcher('levenshtein', lambda **kwargs: TextDistanceMatcher(by_words=False, metric='levenshtein', **kwargs))
register_matcher('cosine', lambda **kwargs: TextDistanceMatcher(by_words=True, metric='cosine', **kwargs))
register_matcher('tf-idf', lambda **kwargs: TFIDFMatcher(**kwargs))
register_matcher('simple_text', lambda weights=(0.5, 0.5), **kwargs: WeightedAverageMatcher([
    TextDistanceMatcher(by_words=False, metric='levenshtein'),
    TextDistanceMatcher(by_words=True, metric='cosine'),
], weights=weights, **kwargs))



import dialogic

from dialogic.testing.testing_utils import make_context


DEFAULT_MESSAGE = 'this is the default message'


class CheckableFormFiller(dialogic.dialog_manager.form_filling.FormFillingDialogManager):
    SIGNS = {
        'jan': 'The Goat',
        'feb': 'The Water Bearer',
        'mar': 'The Fishes',
        'apr': 'The Ram',
        'may': 'The Bull',
        'jun': 'The Twins',
        'jul': 'The Crab',
        'aug': 'The Lion',
        'sep': 'The Virgin',
        'oct': 'The Balance',
        'nov': 'The Scorpion',
        'dec': 'The Archer',
    }

    def handle_completed_form(self, form, user_object, ctx):
        response = dialogic.dialog_manager.base.Response(
            text='Thank you, {}! Now we know: you are {} years old and you are probably {}. Lucky you!'.format(
                form['fields']['name'],
                2019 - int(form['fields']['year']),
                self.SIGNS[form['fields']['month']]
            )
        )
        return response


def test_form():
    dm = CheckableFormFiller(
        'tests/test_managers/form.yaml',
        default_message=DEFAULT_MESSAGE
    )
    resp = dm.respond(make_context(new_session=True))
    assert resp.text == DEFAULT_MESSAGE

    for q, a in [
        ('start the test', 'Please tell me your name'),
        ('Bob', 'Now tell me the year of your birth. Four digits, nothing more.'),
        ('not', 'Please try again. Your answer should be 4 digits.'),
        ('0', 'Please try again. Your answer should be 4 digits.'),
        ('1999', 'Wonderful! Now choose the month of your birth (the first 3 letters).'),
        ('lol', 'The answer should be one of the suggested options - the first 3 letters of a month.'),
        ('jan', 'That\'s great! Finally, tell me the date of your birth - one or two digits'),
        ('40', 'Please try again. Your answer should be a whole number - the day of your birth.'),
        ('02', 'Thank you, Bob! Now we know: you are 20 years old and you are probably The Goat. Lucky you!'),
        ('Okay, what\'s next?', DEFAULT_MESSAGE),
        ('But really', DEFAULT_MESSAGE)
    ]:
        resp = dm.respond(make_context(text=q, prev_response=resp))
        assert resp.text == a, 'expected "{}" to be responded by "{}", got "{}" instead'.format(q, a, resp.text)
